.\"
.\" ${TROFF} ${MSMACROS} thisfile | ${PRINTER}
.\"
.\" revision date - change whenever this file is edited
.ds RD 29 May 1992
.nr PO 1.2i	\" page offset 1.2 inches
.nr PD .7v	\" inter-paragraph distance
.\"
.EH 'troffcvt \*- a troff converter'- % -''
.OH ''- % -'troffcvt \*- a troff converter'
.OF 'Revision date:\0\0\*(RD''Printed:\0\0\n(dy \*(MO 19\n(yr'
.EF 'Revision date:\0\0\*(RD''Printed:\0\0\n(dy \*(MO 19\n(yr'
.\"
.\" I - italic font (taken from -ms and changed)
.de I
.nr PQ \\n(.f
.if t \&\\$3\\f2\\$1\\fP\&\\$2
.if n .if \\n(.$=1 \&\\$1
.if n .if \\n(.$>1 \&\\$1\c
.if n .if \\n(.$>1 \&\\$2
..
.de St	\" troffcvt special text
\\&\\$3\fB@\\$1\fR\\$2
..
.de Cl	\" troffcvt control
\\&\\$3\fB\e\\$1\fR\\$2
..
.de Rq	\" troff request
\\&\\$3\fB\.\\$1\fR\\$2
..
.de Es	\" troff escape
\\&\\$3\fB\e\\$1\fR\\$2
..
.de Ac	\" action
.LP
.ta \n(LLuR
\(bu \\$1	(\fB\\$2\fR)
.br
..
.TL
troffcvt \*- A troff converter
.AU
Paul DuBois
dubois@primate.wisc.edu
.AI
Wisconsin Regional Primate Research Center
Revision date:\0\0\*(RD
.AB
Yet another
.I troff -related
program, with yet another set of misassumptions about how
.I troff
interprets its input, and its own set of deficiencies and bugs.
.AE
.NH
Introduction
.LP
This document describes
.I troffcvt
(``\fItroff\fR convert''), a program which assists in the process of converting
.I troff
documents to other formats.
.I troffcvt
doesn't do the full job of translation itself; rather, it is a preprocessor
that turns
.I troff
files into an intermediate format which is easier to interpret than raw
.I troff .
.I troffcvt
is
intended as a front end that supplies input to a postprocessor which
finishes the translation to produce output in the target format.
Since the job of writing a translator for a given target
format then need not include writing a
.I troff -parser,
the burden of the translator writer is reduced.
In a sense,
.I troffcvt
is simply another sort of
.I ditroff ,
one which produces a different output language.
.LP
.I troffcvt
started out as a
.I troff -to-RTF
converter (a
.I sed
script), but it became evident that this wasn't going to be a
very simple job to do correctly.
It seemed a better justification of effort
to write a more general tool that would be useful in contexts other than
that of RTF production.
The source distribution contains some example
translation methods (simple postprocessors) you can look at.
A standard reader for
.I troffcvt
output is included which can be configured for use with your own
postprocessors.
.LP
.I troffcvt
has a number of significant shortcomings.
It doesn't do very well with input that has been passed through
.I tbl ,
.I eqn
or
.I pic .
.I troff
constructs involving determination of motion or sizes sometimes are
calculated inaccurately since
.I troffcvt
knows nothing of font metrics.
Conditional construct processing is problematic also, as are position-dependent
traps.
.I troffcvt
has other limitations; if those listed here are
insufficient to dissuade you from using it as the basis for a
translator, see the document
.I "``troffcvt \*- Notes, Bugs, Deficiencies''" .
.NH
Translation Model
.LP
.I troff
files consist of text to be formatted interspersed with markup
requests indicating how the formatting is to be done.
The language implemented by
.I troff
is essentially an inverted programming language, where document text
comprises the comments and markup requests are a program
indicating how to format the comments.
This language isn't especially easy to parse, which may be why there
are few tools for translating
.I troff
documents into other formats.
The purpose of
.I troffcvt
is to make it easier to write
.I troff -to-XXX
translators, for arbitrary XXX, by doing
the hard work of turning
.I troff
input into something easier to interpret.
This means part of the job is already done for
postprocessor writers, who can then concentrate on producing
output in the desired target format rather than on figuring out
how to understand
.I troff
files.
.LP
For example, point size might be set by some disgusting sequence like this:
.LP
.DS
\&.ds x *a
\&.ds y *b
\&.nr \e(\e*x\e(\e*x 12
\&.ds \e(\e*y\e(\e*y \e\en(\e(\e*x\e(\e*x
\&.ps \e*(\e(\e*y\e(\e*y
.DE
.LP
This would be digested by
.I troffcvt
and appear in the output in a somewhat simpler form:
.LP
.DS
\epoint-size 12
.DE
.LP
.I Caveat:
The
quality of translators obviously depends on the quality of
.I troffcvt 's
preprocessing, which is suspect.
Nor is the situation improved by the fact that various versions of
.I troff
will do different things with identical input sometimes.
This makes it difficult for
.I troffcvt
to do the ``correct'' thing in all cases, especially for
input files that have been tailored to work with (i.e., around bugs in)
a particular version of
.I troff .
.LP
.I troffcvt
produces output which preserves
information about the structure of a document (e.g., margins, page
length) and its contents (the text it contains).
The goal is
.I not
to lay out text on pages.
That is left to postprocessors, which are expected to lay out document
content by interpreting structure information.
Postprocessors may use the structure and/or content to varying extents.
For instance, a postprocessor which simply extracts the text would ignore
the structure information.
A postprocessor which produces a summary of the structure
(e.g., page layout information), would ignore the text.
Most postprocessors will fall somewhere between these extremes.
.LP
Inevitably, a certain amount of information is lost.
Usually this results from not knowing all the characteristics of
the output device.
For instance, no font metric information is used, so it's not possible
to determine the position on the current page, or even what the current page
number is.
.LP
.DE
An example of
.I troffcvt
operation is shown below.
(A resolution of 432 is assumed.)
.LP
.DS
.ta 3i
.B "Input	Output"

\&.ps 14	\epoint-size 14
\&.vs 16	\espacing 96
\&.ce	\ecenter	
\&.ft B	\efont B	
troffcvt\e\-a troff converter	troffcvt	
\&.ft	@minus	
	a troff converter	
	\ebreak	
	\eadjust-full	
	\efont R	
.DE
.NH
troffcvt Output
.LP
.I troffcvt
produces a mixture of control and text lines.
Control lines correspond to document structure.
Text lines correspond to document content.
Control lines consist of a backslash ``\e'' followed by a control
word and possibly some parameters for the control word,
e.g.,
.Cl space ,
.Cl "font R" ,
.Cl "page-length 4752" .
Text lines are either plain text
written literally to the output,
or begin with a ``@'' to indicate special characters
(for instance,
.St bullet
for the ``\(bu'' character or
.St alpha
for ``\(*a'').
.LP
None of the control or special-text keywords overlap, but it's still
convenient to use different
leading characters ``\e'' and ``@'' to make it easier for simple filter
programs to distinguish between them.
``sed \-e "/\|\e\e\|/d"'' strips control lines, for example.
.LP
.I troffcvt
output is rife with
.I troff -isms,
such as
.Cl need ,
.Cl embolden .
Little effort was made to map these to more general document layout
concepts since it's not clear that there would be any gain in doing so.
.NH
How troffcvt Works
.LP
There are basically two steps to turning
.I troff
files into some other format:
(i) Run
.I troffcvt
to turn
.I troff
files into input for your postprocessor.
This is essentially a matter of configuring its handling of
.I troff
files by specifying an action file, which is covered below.
(ii) Run the postprocessor to generate final output.
Writing
.I troffcvt
postprocessors can be considered separately from understanding
how
.I troffcvt
itself works, so it is covered in a separate document, ``\fItroffcvt
Output Format and Postprocessor Writing.''
.LP
Probably the easiest way to get some idea of the relationship of
.I troffcvt 's
input and output is to run some
.I troff
files through it and look at what comes out.
But to do that, you need an action file.
This is a plain text file you can create with any text editor,
containing lists of symbolic actions used to specify what happens when
requests occur, define special characters and set processing parameters.
.LP
When
.I troffcvt
runs, it
reads one or more action files to configure itself, then processes input
files according to the information in the action files.
.I troffcvt
doesn't have any built-in knowledge about any request.
Stated another way,
.I troffcvt
by default ignores all requests.
It also knows about very few of the characters with special meaning
(by design, since these vary from
.I troff
to
.I troff ).
All of this stuff has to be specified in an action file.
It isn't as onerous to write an action file as it might sound, since you
can use the supplied file as a basis and modify it as necessary
for your own purposes.
.LP
There are two important implications of specifying
operations that
.I troffcvt
should perform
in terms of symbolic actions rather than hardwiring them into the code:
.IP (i)
It allows more flexibility since
.I troffcvt 's
initial state and response to requests can be modified without changing
.I troffcvt
itself.
(Of course this makes it more work to set up in the first place.)
.IP (ii)
Action lists can be fed into the interpreter at times other than the
occurrence of a request in the input.
This is how the list of ``built-in'' special characters is created.
Different versions of
.I troff
often know about different sets of special characters; building the list
at runtime allows different versions to be accommodated.
The initial page layout can also be specified this way,
since although
initial values for processing parameters are the same as those given
in the Ossanna manual, you can change them.
Thus you can set up layouts for letter size, legal, A4, etc.
.LP
Action lists are specified on a line consisting of a leading keyword to
indicate the action list
type (\fBimm\fR or \fBreq\fR, followed by zero or more actions,
to the end of the line.
Action lists can be executed immediately at the time the action file is read,
or they can be associated with requests, to be executed when a given
request occurs in the input.
Examples:
.LP
.DS
imm point-size 10 spacing 12p
imm font R

req ps parse-absrel-num x point-size eol point-size $1
req ce parse-num x eol center $1
req ll parse-absrel-num m line-length eol line-length $1
req ds parse-name parse-string-value y eol define-string $1 $2
req so parse-filename eol push-file $1
.DE
.LP
.B imm
lines are used to configure
.I troffcvt 's
initial state before it starts processing any input files.
.B req
lines tell
.I troffcvt
what to do when requests occur while processing the input.
.LP
Immediate action lists consist of the word
.B imm
followed by a list of actions that are executed as
soon as they are read.
Since this happens before any input is available, the only actions allowed
are those which do not read input (e.g., parsing actions,
.B define-macro )
or generate input (e.g.,
.B push-string ).
.LP
Request action lists are similar but specify a request name, and the
action list has three distinct sections:
.LP
.DS
req \fIrequest-name parsing-actions\fR eol \fInon-parsing-actions ...\fR
.DE
.LP
.I request-name
is the name of the request (without the leading period).
The
.I parsing-actions
section specifies how to parse any request arguments expected.
If it's empty, it means no request arguments are expected (or that
you just want to ignore them.)
The
.B eol
section is mandatory and consists of just that action, to cause
.I troffcvt
to skip to the end of the request line.
(If this were not done, part of the request line
would be read as a separate line to be processed.)
The
.I non-parsing-actions
section specifies what should happen as a result of the request
occuring in the input.
Typically this involves interpreting the request arguments.
If the
.I non-parsing-actions
section is empty, nothing is done with the request.
This is equivalent to not defining any
action list for the request, so the request will simply be ignored
when it occurs.
.LP
.I troffcvt
begins processing the action list for a request as soon as the request is
recognized (and before the rest of the request line has been read).
Internally,
.I troffcvt
associates each action name with the number of arguments that should
follow it in action lists and a function used to implement the
action.
When an action function is executed, it is passed any arguments specified
in the action list.
For instance, the
.Rq so
request can be described
as ``req so parse-filename eol push-file $1''.
The
.B parse-filename
action parses the line on which the request occurs to find a filename,
puts it in
.B $1 ,
and skips to the end of the request line.
.B push-file
pushes the current input file on the file stack and makes the file whose
name is the value of
.B $1
the current file.
Since
.B $1
refers to the first argument parsed from the
.Rq so
request,
if the request is ``.so junk'', then ``push-file $1'' becomes
``push-file junk'',
making ``junk'' the current input file.
.LP
The
.B "req ps"
line above defines the actions to take when a
.Rq ps
request occurs:
parse a number, which can be an absolute setting or relative to the current
point size; skip to the end of the
request line; set the point size using the previously parsed number
($\fIn\fR refers to the \fIn\fR-th request argument parsed earlier
in the action list.)
The
.B "req ce"
line parses a number, skips to the end of the
request line, and causes the next ``$1'' input lines to be centered.
.LP
The language implemented by
.I troff
is expressive (if somewhat unwieldy), so a large number of actions seem to
be necessary to allow requests to be specified properly.
Descriptions of the full set of available actions are available in
``\fItroffcvt Action Reference\fR''.
.NH
Macro Packages
.LP
.I troff
is usually invoked with a
.B \-m \fIxx\fR
flag (e.g.,
.B \-man ,
.B \-me ,
.B \-mm ,
.B \-ms ).
A macro package can be handled at least four ways.
.IP (i)
Ignore it.
Since none of the macros referenced in the input file will then have any
definitions as far as
.I troffcvt
is concerned, they'll all be ignored.
This is probably not very useful.
.IP (ii)
Read the contents of the macro file first before reading the input file named
by the user.
This capability is built into
.I troffcvt :
if an argument
.B \-m \fIxx\fR
is specified,
.I troffcvt
looks for a file
.I /usr/lib/tmac/tmac.xx
and reads it.
.IP
The advantage of this approach is that you don't have to know anything
about how the macro package works.
There are two disadvantages.
First, you may quickly discover just how unsophisticated
.I troffcvt
really is.
Macro packages tend to be very ugly, i.e., clever, complex, powerful,
forbidding, and obviously written by people who understand
.I troff
better than I.
It's not unlikely at all that
.I troffcvt
will botch some of the macros.
Second,
.I troffcvt
may write lots of output in response to a macro invocation in the input.
A paragraphing macro may set point size, vertical spacing, font, indent,
centering, underlining, etc.
That's OK for preserving the appearance of a document.
It won't be helpful if instead you want your postprocessor to recognize
when a macro was used in the
.I troff
file with the idea of creating some sort of corresponding effect
in the target format
(e.g., a style in a
stylesheet), because what you'll see in the
.I troffcvt
output are the effects of all the requests to which the macro expands.
.IP (iii)
List macros as fake requests in an action file.
For instance, the
.B \-ms
macro
.Rq LP
can be described as:
.LP
.DS
req LP eol control-output "LP"
.DE
.IP
This will cause
.Cl LP
to be written to the output when instance of
.Rq LP
occur in the input.
A postprocessor can easily recognize
.Cl LP
and do something sensible.
.IP (iv)
List macros as fake requests in an action file, but define them such
that they mimic the effect of the real macros.
For example, the
.Rq AB
macro from
.B \-ms
can be defined, roughly, as
.LP
.DS
req AB parse-macro-args eol push-string ".ce\en\e\efIABSTRACT\e\efR\en.sp\en.fi\en"
.DE
.IP
You need to understand something about how the macros are supposed to work
for this approach to be fruitful.
(And if you do, you'll see why the
.B req
line just above is deficient.)
.NH
Input/Output Mechanisms
.NH 2
Input Processing
.LP
Input may come from a file, a macro, a named string (created with the
.B define-string
action, usually in response to a
.B \.ds
request), or an anonymous string.
The bulk of input usually comes from
input files named on the command line, which are processed in sequence.
Inputs sources may be nested (e.g., a macro or string
may be referenced while reading a file).
The current input is suspended when another input source is interpolated
into the input stream, and is resumed when the interpolated source is
exhausted.
.LP
.I ChIn()
returns the next input character from the input stream.
Embedded newlines (introduced with a ``\e'' at the end of a line) are deleted
so that the following input line appears contiguous with the current line
to any higher-level routines.
Comments (introduced with \e") are deleted up to (but not including) the
end of line character.
For instance, this makes ``text followed by comment\e" this is the comment''
appear to be ``text followed by comment''.
The handling of lines that begin with \fB\.\e"\fR happens properly since the
comment stripping makes the line look like a line beginning with a control
character but no request, so it is ignored.
.I ChIn()
also manages encoding
of escaped characters, and pushing to input sources for number register,
string or macro argument references.
Handling of escape sequences differs depending on whether
copy mode is in effect or not.
.LP
Input characters accepted by the file-input routine are non-null ASCII
values (null bytes and bytes with bit 8 on are discarded).
This is the range [1..0x7f].
When a character is escaped by preceding it with `\e`, the escape+char
combination is converted internally by
.I ChIn()
to a single value, the ASCII value
of the character with the 8th bit turned on (i.e., \fIc\fR\(bv0x80).
Escape codes are in the range [0x81..0xff], which is distinct from
the range of valid input characters.
.LP
Input source pushing occurs automatically in
.I ChIn()
when
.Es n ,
.Es *
or
.Es $
occur
(and also
.Es w
if not in copy mode).
Higher level routines also can cause the current input source to be pushed
down, e.g., when a
.Rq so
request occurs.
.LP
.I ChIn()
is also used for the ugly task of processing multi-line conditional
input (bracketed with the
.Es {
and
.Es }
sequences).
The conditional request processor saves the current if-level
when it sees a
.Es { ,
bumps it up one,
then processes lines until the level drops back down to the saved
value.
.I ChIn()
notices
.Es } ,
silently deletes it and decrements the if-level, which is then noticed
by the conditional processor.
Pretty horrid.
.LP
.I UnChIn()
is used to push characters back onto the input stream.
It understands how to push back escape codes properly.
It also understands how to push back multiple characters (they
should be pushed in the reverse order from that in which they were read).
.LP
.I ChIn0()
returns the next raw (uninterpreted) character from the input stream.
If there are any pushed back characters waiting to be reread, it returns
the one most recently pushed.
Otherwise, it
returns the next input character from the current input source.
If that source is exhausted, it resumes reading from the previous source.
When there is no more input, it returns EOF.
Input source unwinding is undetectable at any level above
.I ChIn0() ,
including
.I ChIn() .
.LP
.I FChIn() ,
.I MChIn() ,
.I SChIn()
and
.I AChIn()
are the lowest level input routines.
Each returns a single character from a file,
macro,
named string,
or ``anonymous''
string input source.
Each returns EOF when the source is exhausted (which only means the current
source is done, not necessarily that all sources are done).
.LP
.I FChIn()
discards nulls and non-ASCII characters (characters with
the high bit set).
(It also converts CR or CRLF to LF, which has nothing to do with
.I troff ,
but allows text files from MS-DOS or Macintosh
machines to be read without requiring you to convert line endings first.)
.LP
.I MChIn()
reads the next character from a macro definition.
.LP
.I SChIn()
reads the next character from a named string.
.LP
.I AChIn()
reads the next character from an anonymous string,
which is just some arbitrary string that is to be used as an
input source.
For instance, when a number register reference
.Es n ) (
or width expression
.Es w ) (
occur, the resulting value is converted
to a character string, which becomes the current input until
the string is completely read.
References to macro arguments are treated similarly; the argument value
is retrieved and pushed on the input stack.
The
.B push-string
action can be used in action files to push an arbitrary string onto
the input stream, which is convenient for processing certain requests.
.LP
One sticky problem occurs with the
.B \.nx
request, usually processed with the
.B switch-file
action.
When
.B \.nx
occurs, it might happen while other files or macros are active.
If the current input source is a file there is no problem since the file pointer
for that source is simply switched to the new file.
But if the request occurs in the middle of a macro, it's less clear what should
happen.
Should the macro continue to be processed?
I elect to terminate macro sources and unwind the source stack until a file is
found, then switch the file pointer of the file source.
Possibly this is wrong; the
.I troff
manual is ambiguous on this point.
(Which may be why different versions of
.I troff
behave differently in this situation.)
.LP
For handling the
.B \.ex
request, the
.B end-input
action sets a flag causing
.I ChIn ()
to return EOF forever after.
.NH 2
Output Processing
.LP
At the lowest output level, there are two calls.
One is for writing characters and it simply writes to the output
file and dies if there was an error.
The other is for writing strings; it calls the write-character routine
for each character in the string.
.LP
The next level up manages the mechanics of collecting plain text lines and
interspersing them with special text and control lines.
The basic issues are making sure that special text and
control lines don't get written into the middle of a plain text line and
insertion of spaces between successive output text lines.
.LP
Control lines begin with ``\e''.
Any plain text output line being collected is flushed so the control
string doesn't appear on the same line.
.LP
There are two kinds of text output, plain text lines and special text lines
to indicate special characters (e.g.,
.St backslash
for ``\e'').
Whenever text output is written (either plain or special), a check is made
to see whether it's necessary to write a preceding space first.
A space is usually needed between consecutive input text lines (exceptions are
when centering or no-fill are in effect, or if an input line ends with a
.Es c ).
For special text,
any plain text output line being collected is flushed so the control
string doesn't appear on the same line.
.LP
The output character set for text is such that most printable ASCII
characters appear as themselves, and others are written out as special
text lines.
The characters tab, backspace, ``\e'' and ``@'' are printable but
written as specials
.St tab ,
.St backspace ,
.St backslash
and
.St at .
The leader character SOH is written as
.St leader .
.NH
Character Coding
.LP
.I ChIn()
returns an
.I int
value which falls into the following ranges:
.IP "1..127 (0x01..0x7f)"
.br
Plain ASCII character.
.IP "129..255 (0x81..0xff)"
.br
Escape code for character preceded by an escape character (except
.Es ( ,
see below).
The code for \fB\e\fIX\fR is constructed as 0x80\||\|\fIX\fR.
Note that 0x80 is not a valid escape code because null bytes are
stripped from the input.
.IP ">255 (>0xff)"
.br
Special character code.
Sequences of the form \fB\e(\fIxx\fR are recognized and collapsed into
codes of
the form 0x8000\||\|\fIindex\fR, where
.I index
is the index of special character
.I xx .
Index values are in the range 1..255 for reason explained below.
This means only 255 special characters are allowed.
When
.I xx
is not recognized (i.e., not defined in the action file),
a new special character with an empty value is created on the fly.
This is done on the grounds that (a) it's undesirable to simply panic; (b)
it's undesirable to delete the sequence from the input, since that would
break fun constructions like ``\fB\.if
\e(xx\fIstring\fB\e(xx\fIstring\fB\e(xx\fR \fIstuff\fR\ ...''.
A warning is generated when characters are created this way so it can be
known there is a special character definition missing from the action file.
.LP
.I UnChIn()
takes an
.I int
argument, which will usually be a value returned from
.I ChIn() .
.I UnChIn()
pushes the argument onto the input pushback stack, unpacking escape and
special character codes into their original 2- or 4-byte input sequences.
Unpacking this way avoids the problem that can occur when a character is
read in non-copy-mode, pushed back, and read in copy mode.
.LP
Plain ASCII values and escape codes can be represented in a single byte
(as an unsigned character), but special character codes as used with
.I ChIn()
and
.I UnChIn()
are a problem in that they cannot be put into strings directly.
To get around this, special character codes are unpacked into two bytes.
Byte 1 is 0x80 (which is not an otherwise used escape code, since it would
correspond to the null character).
Byte 2 is the special character index, 1..255.
Index 0 is not used because that would translate into a null byte in the
middle of strings, which would, e.g., cause failure of
.I strlen()
and
.I strcmp() .
.LP
Special characters are disallowed in request arguments and escape sequences
that might be written back out directly.
For instance,
.Rq ft
.I F
is written out as
.Cl font
.I F ,
so
.I F
isn't allowed to contain special characters.
Similarly for diversion names.
.LP
Special character names are not allowed to be composed of other special
characters, e.g.,
.B \e(\e(ts\e(ts
is disallowed.
